@@ -33,6 +33,9 @@
  * SUCH DAMAGE.
  */
 
+/* Debug Mode */
+//#define __DEBUG__ 
+
 #include <string.h>
 #include <ctype.h>				/* isspace() */
 #include <getopt.h>
@@ -42,9 +45,6 @@
 #include "libpq-fe.h"
 #include "stack/stack.h"
 
-/* Debug Mode */
-//#define __DEBUG__ 
-
 /* skip space symbol ? -> affect to XPath position() function */
 //#define STRIP_SPACE 
 
@@ -54,6 +54,14 @@
 #define BUFSIZ 1048576		// 1M
 //#define BUFSIZ 2097152    // 2M 
 
+#ifdef _Bool
+#define bool _Bool
+#else
+#ifndef C_H
+typedef char bool;
+#endif//C_H
+#endif//_Bool
+
 #define __DBNAME "xpsql"
 #define __HOST NULL
 #define __PORT NULL
@@ -80,14 +88,7 @@
 
 #define TRANZ_COMMIT		res = PQexec(conn, "COMMIT");			 \
 							PQclear(res)
-
-#define EXIT_STARTELEMENT	free(p);								\
-							free(a1);free(a2);						\
-							g_free(CurrentNode);					\
-							g_string_free(query, TRUE);				\
-							PQclear(res);							\
-							EXIT_DISCONN()
-
+							
 /**
  * ############################################################
  * Memory management - we make expat use standard pg MM
@@ -126,18 +127,20 @@ pgxml_mhs_init()
 
 // ############################################################
 
-_Bool strip_space;
+bool strip_space;
 static PGconn *conn;
 static signed int CurrentParent;	// int32
-static _Bool StackIsLoaded;
+static bool StackIsLoaded;
 static Stack *textNodeStack;
-static _Bool StackNsIsLoaded;
+static bool StackNsIsLoaded;
 static Stack *nsprefixStack;
 static Stack *nsuriStack;
 
 char *from_codeset;
 char *to_codeset;
 
+char buf[MAX_CONTENT_SIZE];
+
 /**
  * expat
  */
@@ -151,12 +154,12 @@ void xmlDecl(void *, const XML_Char *, const XML_Char *, int);
 //void endNS(void *, const XML_Char *); 
 void defaultData(void *, const XML_Char *, int);
 
-static void trim(char *);
+//static void trim(char *);
 static char *strcompress(const char *);
-static int StrReplace(char *, const char *, const char *);
+//static int StrReplace(char *, const char *, const char *);
 static char *basename(char *);
-static _Bool setTextNode(void);
-//static _Bool setNsNode(void); 
+static bool setTextNode(void);
+//static bool setNsNode(void); 
 static void showUsage();
 
 // ############################################################
@@ -164,16 +167,13 @@ void
 startElement(void *userData, const char *name, const char **attrs)
 {
 	int i = 0;
-	char *att_id;
+	char *att_id, *tag_name;
+	char *att_name, *att_value;
+	char *a1, *a2;
 	char *CurrentNode;
-	char *tag_name = (char *) malloc(strlen(name) + 1);
-	char *p = tag_name;
-	char *att_name = (char *) malloc(MAX_CONTENT_SIZE + 1);
-	char *a1 = att_name;
-	char *att_value = (char *) malloc(MAX_CONTENT_SIZE + 1);
-	char *a2 = att_value;
 	PGresult *res;
 	GString *query = g_string_sized_new(100);
+	att_name = att_value = NULL;
 
 	if (StackIsLoaded)
 		setTextNode();
@@ -181,10 +181,10 @@ startElement(void *userData, const char *name, const char **attrs)
 	/**
 	 * convert UTF8 to EUC
 	 */
-	memcpy(tag_name, name, strlen(name) + 1);
-	g_strstrip(tag_name);
+	memcpy(buf, name, strlen(name) + 1);
+	g_strstrip(buf);
 	tag_name =
-		g_convert(tag_name, strlen(tag_name), to_codeset, from_codeset,
+		g_convert(buf, strlen(buf), to_codeset, from_codeset,
 				  NULL, NULL, NULL);
 
 #ifdef __DEBUG__
@@ -195,20 +195,23 @@ startElement(void *userData, const char *name, const char **attrs)
 	/**
 	 * create element
 	 */
-	g_string_sprintf(query, "SELECT new_childelement(%d,'%s')",
+	g_string_sprintf(query, "SELECT addchildelement(%d,'%s')",
 					 CurrentParent, tag_name);
 
 #ifdef __DEBUG__
-	fprintf(stdout, "SELECT new_childelement(%d,'%s')\n", CurrentParent,
+	fprintf(stdout, "SELECT addchildelement(%d,'%s')\n", CurrentParent,
 			tag_name);
 #endif
 
 	res = PQexec(conn, query->str);
 	CurrentNode = g_strdup(PQgetvalue(res, 0, 0));
 	if (!res || PQresultStatus(res) != PGRES_TUPLES_OK || CurrentNode == NULL) {
-		fprintf(stderr, "Error\tnew_childelement(%d,%s)\n", CurrentParent,
+		fprintf(stderr, "Error\taddchildelement(%d,%s)\n", CurrentParent,
 				tag_name);
-		EXIT_STARTELEMENT;
+		g_free(tag_name);
+		g_string_free(query, TRUE);
+		PQclear(res);
+		EXIT_DISCONN();
 	}
 	PQclear(res);
 
@@ -230,18 +233,15 @@ startElement(void *userData, const char *name, const char **attrs)
 	 * set attribute
 	 */
 	while (attrs[i] != '\0') {
-		att_name = a1;
-		strcpy(att_name, attrs[i]);
-		g_strstrip(att_name);
 		att_name =
-			g_convert(att_name, strlen(att_name), to_codeset, from_codeset,
+			g_convert(attrs[i], strlen(attrs[i]), to_codeset, from_codeset,
 					  NULL, NULL, NULL);
-		att_value = a2;
-		strcpy(att_value, attrs[i + 1]);
-		g_strstrip(att_value);
+		a1 = g_strstrip(att_name);
+		PQescapeString(buf, attrs[i + 1], strlen(attrs[i + 1]));
 		att_value =
-			g_convert(att_value, strlen(att_value), to_codeset,
-					  from_codeset, NULL, NULL, NULL);
+			g_convert(buf, strlen(buf), to_codeset, from_codeset, 
+					NULL, NULL, NULL);
+		a2 = att_value;
 
 		if (strstr(att_name, "xmlns:") != NULL) {
 			att_name = att_name + 6;
@@ -252,7 +252,12 @@ startElement(void *userData, const char *name, const char **attrs)
 			}
 			else
 			{
-				EXIT_STARTELEMENT;
+				g_free(tag_name);
+				g_free(a1);
+				g_free(a2);
+				g_string_free(query, TRUE);
+				PQclear(res);
+				EXIT_DISCONN();
 			}
 		}
 		else if (strstr(att_name, "xmlns=") != NULL) {
@@ -269,7 +274,12 @@ startElement(void *userData, const char *name, const char **attrs)
 		att_id = PQgetvalue(res, 0, 0);
 		if (!res || PQresultStatus(res) != PGRES_TUPLES_OK || att_id == NULL) {
 			fprintf(stdout, "error is occuring at...\n%s\n", query->str);
-			EXIT_STARTELEMENT;
+			g_free(tag_name);
+			g_free(a1);
+			g_free(a2);
+			g_string_free(query, TRUE);
+			PQclear(res);
+			EXIT_DISCONN();
 		}
 		PQclear(res);
 
@@ -277,6 +287,9 @@ startElement(void *userData, const char *name, const char **attrs)
 		fprintf(stdout, "create attribute ...name='%s' value='%s'\n",
 				attrs[i], attrs[i + 1]);
 #endif
+		g_free(a1);
+		g_free(a2);
+
 		i += 2;
 	}
 
@@ -285,9 +298,7 @@ startElement(void *userData, const char *name, const char **attrs)
 	 */
 	g_string_free(query, TRUE);
 	g_free(CurrentNode);
-	free(p);
-	free(a1);
-	free(a2);
+	g_free(tag_name);
 }
 
 void
@@ -331,9 +342,9 @@ defaultData(void *userData, const XML_Char * s, int len)
 void
 commentData(void *userData, const XML_Char * data)
 {
-	char *comment = (char *) malloc(strlen(data) + 1);
-	char *p = comment;
+	char *comment=NULL;
 	PGresult *res;
+	GError *err = NULL;
 	GString *query = g_string_sized_new(100);
 
 	if (StackIsLoaded)
@@ -342,15 +353,22 @@ commentData(void *userData, const XML_Char * data)
 	/**
 	 * convert UTF8 -> EUC
 	 */
-	memcpy(comment, data, strlen(data) + 1);
+	PQescapeString(buf, data, strlen(data));
 	comment =
-		g_convert(comment, strlen(comment), from_codeset, to_codeset, NULL,
-				  NULL, NULL);
+		g_convert(buf, strlen(buf), to_codeset, from_codeset, NULL,
+				  NULL, &err);
+	if (err != NULL) {
+		fprintf(stderr, "error:%s\n", err->message);
+		g_error_free(err);
+		g_free(comment);
+		g_string_free(query, TRUE);
+		exit(1);
+	}
 
 #ifdef __DEBUG__
 	printf("[commentData]\n");
 #endif
-	g_string_sprintf(query, "SELECT new_childcomment(%d,'%s')",
+	g_string_sprintf(query, "SELECT addchildcomment(%d,'%s')",
 					 CurrentParent, comment);
 	res = PQexec(conn, query->str);
 	if (!res ||
@@ -358,7 +376,7 @@ commentData(void *userData, const XML_Char * data)
 		PQgetvalue(res, 0, 0) == NULL) {
 		fprintf(stderr, "create comment under element...='%d'\n",
 				CurrentParent);
-		free(p);
+		g_free(comment);
 		g_string_free(query, TRUE);
 		PQclear(res);
 		EXIT_DISCONN();
@@ -366,16 +384,14 @@ commentData(void *userData, const XML_Char * data)
 	PQclear(res);
 
 	g_string_free(query, TRUE);
-	free(p);
+	g_free(comment);
 }
 
 void
 piData(void *userData, const XML_Char * target, const XML_Char * data)
 {
-	char *pitarget = (char *) malloc(strlen(target) + 1);
-	char *pidata = (char *) malloc(strlen(data) + 1);
-	char *p1 = pitarget;
-	char *p2 = pidata;
+	char *pitarget;
+	char *pidata;
 	PGresult *res;
 	GString *query = g_string_sized_new(100);
 
@@ -386,24 +402,23 @@ piData(void *userData, const XML_Char * target, const XML_Char * data)
 	if (StackIsLoaded)
 		setTextNode();
 
-	memcpy(pitarget, target, strlen(target) + 1);
-	memcpy(pidata, data, strlen(data) + 1);
 	pitarget =
-		g_convert(pitarget, strlen(pitarget), from_codeset, to_codeset,
+		g_convert(target, strlen(target), from_codeset, to_codeset,
 				  NULL, NULL, NULL);
+	PQescapeString(buf, data, strlen(data));
 	pidata =
-		g_convert(pidata, strlen(pidata), from_codeset, to_codeset, NULL,
+		g_convert(buf, strlen(buf), from_codeset, to_codeset, NULL,
 				  NULL, NULL);
 
-	g_string_sprintf(query, "SELECT new_childpi(%d,'%s','%s')",
+	g_string_sprintf(query, "SELECT addchildpi(%d,'%s','%s')",
 					 CurrentParent, pitarget, pidata);
 	res = PQexec(conn, query->str);
 	if (!res ||
 		PQresultStatus(res) != PGRES_TUPLES_OK ||
 		PQgetvalue(res, 0, 0) == NULL) {
 		fprintf(stderr, "Can't Create PI ...'%s'\n", pidata);
-		free(p1);
-		free(p2);
+		free(pitarget);
+		free(pidata);
 		g_string_free(query, TRUE);
 		PQclear(res);
 		EXIT_DISCONN();
@@ -415,8 +430,8 @@ piData(void *userData, const XML_Char * target, const XML_Char * data)
 #endif
 
 	g_string_free(query, TRUE);
-	free(p1);
-	free(p2);
+	g_free(pitarget);
+	g_free(pidata);
 }
 
 /*
@@ -462,39 +477,33 @@ xmlDecl(void *userData, const XML_Char * version,
 void
 charData(void *userData, const XML_Char * s, int len)
 {
-	char *str = (char *) malloc(len * 2 + 1);
+	char *str;
 	GError *err = NULL;
 
 #ifdef __DEBUG__
 	printf("[XML_CharacterDataHandler]\n");
 #endif
 
-	if (str != NULL) {
-		PQescapeString((char *) str, (char *) s, len);
-	 	//StrReplace(str,"'","''");        // select new_childText(1,'')
-		StrReplace(str, "\n", "\\n");
-		StrReplace(str, "\t", "\\t");
+	PQescapeString(buf, (char *) s, len);
+	//StrReplace(buf, "\n", "\\n");
+	//StrReplace(buf, "\t", "\\t");
 
-#ifdef __DEBUG__
-		printf("before conversion:%s\n", str);
-#endif
-
-		if (*str != '\0') {
-			str =
-				g_convert(str, strlen(str), to_codeset, from_codeset, NULL,
-						  NULL, &err);
-			if (err != NULL) {
-				fprintf(stderr, "error:%s\n", err->message);
-				g_error_free(err);
-				exit(1);
-			}
-#ifdef __DEBUG__
-			printf("content:%s\n", str);
-#endif
-			stack_push_pointer(textNodeStack, str);
-			StackIsLoaded = TRUE;
+	if (*buf != '\0') {
+		str =
+			g_convert(buf, strlen(buf), to_codeset, from_codeset, NULL,
+					  NULL, &err);
+		if (err != NULL) {
+			fprintf(stderr, "error:%s\n", err->message);
+			g_error_free(err);
+			exit(1);
 		}
+#ifdef __DEBUG__
+		printf("content:%s\n", str);
+#endif
+		stack_push_pointer(textNodeStack, str);
+		StackIsLoaded = TRUE;
 	}
+
 	return;
 }
 
@@ -503,50 +512,50 @@ charData(void *userData, const XML_Char * s, int len)
  *
  *  this function needs care for handring, buffer overflow
  */
-static int
-StrReplace(char *ap, const char *bp, const char *cp)
-{
-	int bs, cs, rc;
-	char *limit;
-
-	rc = 0;
-	bs = strlen(bp);
-	cs = strlen(cp);
-	limit = ap + strlen(ap) - bs;
-
-	while (ap <= limit) {
-		if (strncmp(ap, bp, bs) == 0) {
-			memmove(ap + cs, ap + bs, limit - ap + 1);
-			strncpy(ap, cp, cs);
-			limit += cs - bs;
-			rc++;
-			ap += cs;
-		}
-		else
-			ap++;
-	}
-	return rc;
-}// streplace
+//static int
+//StrReplace(char *ap, const char *bp, const char *cp)
+//{
+//	int bs, cs, rc;
+//	char *limit;
+//
+//	rc = 0;
+//	bs = strlen(bp);
+//	cs = strlen(cp);
+//	limit = ap + strlen(ap) - bs;
+//
+//	while (ap <= limit) {
+//		if (strncmp(ap, bp, bs) == 0) {
+//			memmove(ap + cs, ap + bs, limit - ap + 1);
+//			strncpy(ap, cp, cs);
+//			limit += cs - bs;
+//			rc++;
+//			ap += cs;
+//		}
+//		else
+//			ap++;
+//	}
+//	return rc;
+//}// streplace
 
 /*
  *  trim - Remove leading, trailing, & excess embedded spaces
  */
-static void
-trim(char *s)
-{
-	int len = strlen(s);
-
-	/*
-	 * trim trailing whitespace
-	 */
-	while (len > 0 && isspace(s[len - 1]))
-		s[--len] = '\0';
-
-	/*
-	 * trim leading whitespace
-	 */
-	memmove(s, &s[strspn(s, " \n\r\t\v")], len);
-}// trim
+//static void
+//trim(char *s)
+//{
+//	int len = strlen(s);
+//
+//	/*
+//	 * trim trailing whitespace
+//	 */
+//	while (len > 0 && isspace(s[len - 1]))
+//		s[--len] = '\0';
+//
+//	/*
+//	 * trim leading whitespace
+//	 */
+//	memmove(s, &s[strspn(s, " \n\r\t\v")], len);
+//}// trim
 
 static char *
 strcompress(const char *source)
@@ -599,16 +608,20 @@ basename(char *fn)
 	return fn;
 }//basename
 
-static _Bool
+static bool
 setTextNode(void)
 {
-	GString *query = g_string_sized_new(120);
+	GString *query = g_string_sized_new(128);
 	int i;
 	char *str;
 	char *text = NULL;
 	PGresult *res;
 	char *tmp;
 
+#ifdef __DEBUG__
+	printf("[setTextNode]\n");
+#endif
+
 	for (i = 0; textNodeStack->used > 0; i++) {
 		str = stack_pop(textNodeStack);
 		if (i == 0)
@@ -624,20 +637,20 @@ setTextNode(void)
 		if (text == NULL)
 			return FALSE;
 		tmp = strcompress(text);
-		trim(tmp);
+		//trim(tmp);
 		if (*tmp == '\0')
 			return FALSE;		//equals
 		g_free(tmp);
 	}
 
-	g_string_sprintf(query, "SELECT new_childtext(%d,'%s')",
+	g_string_sprintf(query, "SELECT addchildtext(%d,'%s')",
 					 CurrentParent, text);
 	res = PQexec(conn, query->str);
 	if (!res ||
 		PQresultStatus(res) != PGRES_TUPLES_OK ||
 		PQgetvalue(res, 0, 0) == NULL) {
 		fprintf(stderr, "Could not create TextNode\n");
-		fprintf(stderr, "SELECT new_childtext(%d,'%s')\n",
+		fprintf(stderr, "SELECT addchildtext(%d,'%s')\n",
 				CurrentParent, text);
 		PQclear(res);
 		g_string_free(query, TRUE);
@@ -662,7 +675,7 @@ setTextNode(void)
 }//setTextNode
 
 /*
-   static _Bool setNsNode(void)
+   static bool setNsNode(void)
    {
    int i;
    char *uri,*prefix;
@@ -748,8 +761,8 @@ main(int argc, char *argv[])
 	char *dbname = __DBNAME;
 	char *password = __PWD;
 	char *fname;
-	_Bool has_from = FALSE;
-	_Bool has_to = FALSE;
+	bool has_from = FALSE;
+	bool has_to = FALSE;
 
 #ifdef STRIP_SPACE
 	strip_space = TRUE;
