@@ -45,14 +45,10 @@
 
 //#define __DEBUG__
 #define __PG__
-
 #define GEQO_THRESHOLD 11		/* set the value in postgresql.conf, default 11 */
-#define FORCE_USE_IDX_AXIS 
-
+#define FORCE_USE_IDX_AXIS 		/* should be always setted */
 //#define XPATH2
 
-#define GET_STR(textp) DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp)))
-
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>			/* free() */
@@ -74,23 +70,33 @@
 #include <math.h>			/* fbs() */
 
 Datum xpath_eval(PG_FUNCTION_ARGS);
+Datum xpath_evalx(PG_FUNCTION_ARGS);
 Datum xpath2sql(PG_FUNCTION_ARGS);
 Datum vchar2int(PG_FUNCTION_ARGS);
 Datum vchar2real(PG_FUNCTION_ARGS);
 
 #define FLOAT4_MAX		 FLT_MAX
 #define FLOAT4_MIN		 FLT_MIN
+#define GET_STR(textp) DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp)))
 
 #ifndef NAN
 #define NAN		(0.0/0.0)
-#endif
+#endif//NAN
 
 static void CheckFloat4Val(double val);
 
-#endif // __PG__
+#endif//__PG__
 
 #define BUFSIZE 1024
 
+#ifdef _Bool
+#define bool _Bool
+#else
+#ifndef C_H
+typedef char bool;
+#endif//C_H
+#endif//_Bool
+
 #define KIND_ROOT       0
 #define KIND_ELEMENT    1
 #define KIND_TEXT       2
@@ -117,7 +123,7 @@ static void CheckFloat4Val(double val);
 #define RET_ORD_SELECT  0x10			// 1). return select value ordered
 #define RET_FLG		  	0x20			// 2). return node-flagment un-ordered
 #define RET_ORD_FLG	  	0x30			// 3). return node-flagment ordered
-#define RET_XML		  	0x31			// return node-flagment un-ordered
+#define RET_XML		  	0x01			// 4). return xml
 #define SELECT_INTO		0xc0			// 1100 0000
 #define CREATE_VIEW     0x80			// 1000 0000 
 
@@ -173,9 +179,9 @@ typedef struct XPathData_tag
 	//int cnt_pos;				//temp table counter, IOW position() counter
 	int depth;					//path depth
 	unsigned char eval_mode;	//XPath evaluation mode
-	_Bool distinct;				//use dintinct
-	_Bool path_q;				//Expr only with Path
-	_Bool popflag;				//pop stack
+	bool distinct;				//use dintinct
+	bool path_q;				//Expr only with Path
+	bool popflag;				//pop stack
 	GString *spath;				//Path String
 	QStack *qstack_val;			//Query Concreate Stack
 	AxisType evaledAxis;		//current Axis
@@ -187,11 +193,11 @@ static int translateStep(char *, GString **, XPathData *);
 static int translateAxis(char *, GString **, XPathData *);
 static int translateNodeTest(char *, GString **, XPathData *);
 static int translatePredicate(char *, GString **, XPathData *, int);
-static _Bool makeSQL(GString **,XPathData *,int,char);
+static bool makeSQL(GString **,XPathData *,int,char);
 static char *getFormalExp(char *);
 //static int getPathDepth(char *);
 static int StrReplace( char *, const char *, const char *);
-static _Bool StrisDigit(char *);
+static bool StrisDigit(char *);
 char *strrstr(const char *, const char *);
 //int chrcnt(const char *, char);
 static void trim(char *);
@@ -248,10 +254,6 @@ xpath_eval(PG_FUNCTION_ARGS)
 				case 3:
 					mode = RET_ORD_FLG;
 					break;
-				case 4:
-					// not yet supported
-					mode = RET_XML;
-					//break;
 				default:
 #ifdef __DEBUG__
 					elog(ERROR,"mode number is wrong\n");
@@ -378,6 +380,156 @@ xpath_eval(PG_FUNCTION_ARGS)
 }
 
 
+PG_FUNCTION_INFO_V1(xpath_evalx);
+Datum
+xpath_evalx(PG_FUNCTION_ARGS)
+{
+	FuncCallContext *funcctx;
+	int			call_cntr;
+	int			max_calls;
+	TupleTableSlot *slot;
+	AttInMetadata *attinmeta;
+	TupleDesc s_tupdesc;
+	SPITupleTable *fctx_tuptable;
+
+	/* stuff done only on the first call of the function */
+	if (SRF_IS_FIRSTCALL())
+	{
+
+		MemoryContext oldcontext;
+		int ret;
+		char *xpr;
+		char *sql;
+
+#ifdef __DEBUG__
+		elog(NOTICE,"[xpath_eval] enter");
+#endif
+
+		/* create a function context for cross-call persistence */
+		funcctx = SRF_FIRSTCALL_INIT();
+
+		/* switch to memory context appropriate for multiple function calls */
+		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
+
+		xpr = GET_STR(PG_GETARG_TEXT_P(0));
+
+		if( (sql = translateXPath(xpr,RET_XML)) == FALSE){
+
+#ifdef __DEBUG__
+			elog(ERROR,"wrong XPath expression\n");
+#endif
+			PG_RETURN_NULL();
+		}
+
+
+	 	/* Connect to SPI manager */
+	 	if ((ret = SPI_connect()) < 0)
+			elog(ERROR, "eval_sql: SPI_connect returned %d\n", ret);
+
+		/* Retrieve the desired tuple */
+		ret = SPI_exec(sql, 0);
+		pfree(xpr);
+		g_free(sql);
+
+#ifdef __DEBUG__
+		elog(NOTICE,"SPI_processed=%d\n",SPI_processed);
+#endif
+
+		/*
+		 * Only allow one qualifying tuple
+		 */
+		if ((ret == SPI_OK_SELECT) && (SPI_processed > 0)){
+			funcctx->max_calls = SPI_processed;
+			s_tupdesc = SPI_tuptable->tupdesc;
+		}
+		else
+		{
+			SPI_finish();
+			// RESULT SET IS NULL
+			PG_RETURN_NULL();
+		}
+
+		/* allocate a slot */
+		slot = TupleDescGetSlot(s_tupdesc);
+		funcctx->slot = slot;
+
+        /* Generate attribute metadata needed later */
+		attinmeta = TupleDescGetAttInMetadata(s_tupdesc);
+		funcctx->attinmeta = attinmeta;
+
+		/* allocate memory for user context */
+		fctx_tuptable = (SPITupleTable *) palloc(sizeof(SPITupleTable));
+		fctx_tuptable = SPI_tuptable;
+
+		/* Save spi data for use across calls */
+		funcctx->user_fctx = fctx_tuptable;
+
+		MemoryContextSwitchTo(oldcontext);
+	}
+
+	/* stuff done on every call of the function */
+	funcctx = SRF_PERCALL_SETUP();
+
+	/* return slot for our tuple */
+	slot = funcctx->slot;
+	attinmeta = funcctx->attinmeta;
+
+	/* initialize per-call variables */
+	call_cntr = funcctx->call_cntr;
+	max_calls = funcctx->max_calls;
+
+	/* the sql tuple */
+	fctx_tuptable = (SPITupleTable *)funcctx->user_fctx;
+	s_tupdesc = fctx_tuptable->tupdesc;
+
+
+	if (call_cntr < max_calls)	/* do when there is more left to send */
+	{
+		char	  **values;
+		HeapTuple	res_tuple;
+		HeapTuple	s_tuple;
+		Datum		result;
+		int i;
+		int	nfields = s_tupdesc->natts;
+
+		/* get the next sql result tuple */
+		s_tuple = fctx_tuptable->vals[call_cntr];
+
+#ifdef __DEBUG__
+		elog(NOTICE,"nfields=%d\n",nfields);
+#endif
+
+		values = (char **) palloc(nfields * sizeof(char *));
+		for (i = 0; i < nfields; i++){
+			values[i] = SPI_getvalue(s_tuple,s_tupdesc,i+1);
+#ifdef __DEBUG__
+			elog(NOTICE,"call_cntr[%d]:values[%d]=%s\n",call_cntr,i,values[i]);
+#endif
+		}
+
+		/* build the tuple */
+		res_tuple = BuildTupleFromCStrings(attinmeta, values);
+
+		/* make the tuple into a datum */
+		result = TupleGetDatum(slot, res_tuple);
+
+		/* Clean up (this is not actually necessary) */
+		for(i=0;i<nfields;i++){
+			if (values[i] != NULL)
+				pfree(values[i]);
+		}
+		pfree(values);
+
+		SRF_RETURN_NEXT(funcctx,result);
+	}
+	else	/* do when there is no more left */
+	{
+		pfree(fctx_tuptable);
+		SPI_finish();
+		SRF_RETURN_DONE(funcctx);
+	}
+}
+
 PG_FUNCTION_INFO_V1(xpath2sql);
 Datum
 	xpath2sql(PG_FUNCTION_ARGS)
@@ -404,9 +556,8 @@ Datum
 				mode = RET_ORD_FLG;
 				break;
 			case 4:
-				// not yet supported
 				mode = RET_XML;
-				//break;
+				break;
 			default:
 #ifdef __DEBUG__
 				elog(ERROR,"mode number is wrong\n");
@@ -526,12 +677,13 @@ static char *translateXPath(char *argv,char mode)
 	char *bs;
 	char *buf;
 	char sep[] = "/";
-	_Bool justroot;
+	bool justroot;
 	GString *res;
 	XPathData xd = {1,0,0,RET_FLG,FALSE,TRUE,FALSE,NULL,NULL,0};
 	XPathData *xd_head = &xd;
-	xd_head->eval_mode = mode;
 	QStack qs;
+	
+	xd_head->eval_mode = mode;
 	QStack *qs_head = &qs;
 
 #ifdef __DEBUG__
@@ -659,7 +811,7 @@ static int translateStep(char *step, GString **query, XPathData *xd)
 	regex_t preg;
 	char *pattern = "[[:digit:]]+\\]";
 	int i;
-//	_Bool flg = FALSE;
+//	bool flg = FALSE;
 	prevCondition *lprevCond = &(xd->prevCond);
 	QStack *lqs;
 
@@ -779,6 +931,7 @@ static int translateAxis(char *axis, GString **query, XPathData *xd)
 		    	xd->evaledAxis = AXIS_ANCESTOR;
 				xd->distinct = TRUE;
 				xd->path_q = FALSE;
+				// TODO
 				str = "_int_ancestor";
 				break;
 		    }
@@ -787,6 +940,7 @@ static int translateAxis(char *axis, GString **query, XPathData *xd)
 		    	xd->evaledAxis = AXIS_ANCESTOR_OR_SELF;
 		    	xd->distinct = TRUE;
 		    	xd->path_q = FALSE;
+		    	// TODO
 				str = "_int_ancestor_or_self";
 				break;
 		    }
@@ -824,12 +978,12 @@ static int translateAxis(char *axis, GString **query, XPathData *xd)
 					xd->fcount++;
 
 					buf = g_strdup_printf("xn%d.parent = xn%d.id AND "
-										  "xn%d.kind > 0 AND xn%d.kind < 5",
-										  lcnt,lcnt-1,lcnt,lcnt
+										  "xn%d.kind < 5",
+										  lcnt,lcnt-1,lcnt
 										  );
 					stack_push(lqs->ST_Where, buf, strlen(buf)+1);
-
 					g_free(buf);
+					
 					return TRUE;
 				}
 				break;
@@ -839,15 +993,37 @@ static int translateAxis(char *axis, GString **query, XPathData *xd)
 		    xd->distinct = TRUE;
 		    if (!strcmp(axis,"descendant"))
 		    {
-				str = "_int_descendant";
+				buf = g_strdup_printf("xml_node xn%d",lcnt);
+				stack_push(lqs->ST_From, buf, strlen(buf)+1);
+				xd->fcount++;
+				
+				buf = g_strdup_printf("xn%d.dewey > xn%d.dewey AND "
+									  "xn%d.dewey < nextSibling(xn%d.dewey) AND "
+									  "xn%d.kind < 5",
+									lcnt,lcnt-1,lcnt,lcnt-1,lcnt
+									);
+				stack_push(lqs->ST_Where, buf, strlen(buf)+1);
+				g_free(buf);
+				
 				xd->evaledAxis = AXIS_DESCENDANT;
-				break;
+				return TRUE;
 		    }
 		    if (!strcmp(axis,"descendant-or-self"))
 		    {
-				str = "_int_descendant_or_self";
+				buf = g_strdup_printf("xml_node xn%d",lcnt);
+				stack_push(lqs->ST_From, buf, strlen(buf)+1);
+				xd->fcount++;
+				
+				buf = g_strdup_printf("xn%d.dewey >= xn%d.dewey AND "
+									  "xn%d.dewey < nextSibling(xn%d.dewey) AND "
+									  "(xn%d.kind < 5 OR xn%d.id=xn%d.id)",
+									lcnt,lcnt-1,lcnt,lcnt-1,lcnt,lcnt,lcnt-1
+									);
+				stack_push(lqs->ST_Where, buf, strlen(buf)+1);
+				g_free(buf);
+				
 				xd->evaledAxis = AXIS_DESCENDANT_OR_SELF;
-				break;
+				return TRUE;
 		    }
 			return FALSE;
 		case 'f':
@@ -863,9 +1039,11 @@ static int translateAxis(char *axis, GString **query, XPathData *xd)
 					xd->fcount++;
 
 #ifdef FORCE_USE_IDX_AXIS
-					buf = g_strdup_printf("xn%d.dewey < xn%d.dewey AND "
-										  "_int_following(xn%d.dewey,xn%d.dewey)",
-										  lcnt-1,lcnt,lcnt,lcnt-1
+					// Since it is a hedge model, have to insert in between.
+					// xn3.docid = xn2.docid
+					buf = g_strdup_printf("xn%d.dewey >= nextSibling(xn%d.dewey) AND "
+										  "xn%d.dewey > subarray(xn%d.dewey,1,3)",
+										  lcnt,lcnt-1,lcnt,lcnt-1
 										  );
 #else
 					buf = g_strdup_printf("_int_following(xn%d.dewey,xn%d.dewey)",
@@ -890,9 +1068,9 @@ static int translateAxis(char *axis, GString **query, XPathData *xd)
 					xd->fcount++;
 
 					buf = g_strdup_printf("xn%d.parent = xn%d.parent AND "
-										  "xn%d.dewey < xn%d.dewey AND "
-										  "xn%d.kind > 0 AND xn%d.kind < 5",
-										  lcnt,lcnt-1,lcnt-1,lcnt,lcnt,lcnt
+										  "xn%d.dewey > xn%d.dewey AND "
+										  "xn%d.kind < 5",
+										  lcnt,lcnt-1,lcnt,lcnt-1,lcnt
 										  );
 					stack_push(lqs->ST_Where, buf, strlen(buf)+1);
 
@@ -906,7 +1084,7 @@ static int translateAxis(char *axis, GString **query, XPathData *xd)
 		    if (!strcmp(axis,"namespace"))
 		    {
 				xd->evaledAxis = AXIS_NAMESPACE;
-				xd->path_q = FALSE;		// current not supported
+				xd->path_q = FALSE;		// current ot supported
 				if(xd->scount!=0)
 				{
 					buf = g_strdup_printf("xml_node xn%d",lcnt);
@@ -957,6 +1135,7 @@ static int translateAxis(char *axis, GString **query, XPathData *xd)
 					xd->fcount++;
 
 #ifdef FORCE_USE_IDX_AXIS
+					// TODO
 					buf = g_strdup_printf("xn%d.dewey < xn%d.dewey AND "
 										  "_int_preceding(xn%d.dewey,xn%d.dewey)",
 										  lcnt,lcnt-1,lcnt,lcnt-1
@@ -985,8 +1164,8 @@ static int translateAxis(char *axis, GString **query, XPathData *xd)
 
 					buf = g_strdup_printf("xn%d.parent = xn%d.parent AND "
 										  "xn%d.dewey < xn%d.dewey AND "
-										  "xn%d.kind > 0 AND xn%d.kind < 5",
-										  lcnt,lcnt-1,lcnt,lcnt-1,lcnt,lcnt
+										  "xn%d.kind < 5",
+										  lcnt,lcnt-1,lcnt,lcnt-1,lcnt
 										  );
 					stack_push(lqs->ST_Where, buf, strlen(buf)+1);
 
@@ -1070,12 +1249,13 @@ static int translateNodeTest(char *ntest, GString **query, XPathData *xd)
 	char *ptr,*sp;
 	char *ope;
 	char *buf				 = NULL;
-	_Bool lpq				 = FALSE;
+	bool lpq				 = FALSE;
 	NodeTestType lNTest;
 	int lcnt				 = xd->scount;
 	AxisType lAxis			 = xd->evaledAxis;
 	QStack *lqs				 = xd->qstack_val;
 	prevCondition *lprevCond = &(xd->prevCond);
+	bool hasDep = TRUE;	// includes dependant nodes ?
 
 #ifdef __DEBUG__
 	#ifdef __PG__
@@ -1100,7 +1280,7 @@ static int translateNodeTest(char *ntest, GString **query, XPathData *xd)
 	else if(!strcmp(ntest,"node()"))
 	{
 		lNTest = NTest_NODE;
-
+		
 		if(xd->path_q)
 		{
 			switch(lAxis)
@@ -1111,6 +1291,7 @@ static int translateNodeTest(char *ntest, GString **query, XPathData *xd)
 					return TRUE;
 				case AXIS_DESCENDANT:
 					lpq = TRUE;
+					hasDep = FALSE;
 					g_string_append(xd->spath,"#/%");
 					break;
 				case AXIS_ATTRIBUTE:
@@ -1124,30 +1305,13 @@ static int translateNodeTest(char *ntest, GString **query, XPathData *xd)
 					break;
 			}
 		}
+		else
+		{
+			if(lAxis==AXIS_SELF)
+				xd->scount -= 1;
+		}
+
 
-/*
-		if(lAxis==AXIS_DESCENDANT_OR_SELF)
-		{
-	    	xd->path_q = TRUE;
-			lprevCond->prevNTest = lNTest;
-			return TRUE;
-		}
-		if(lAxis==AXIS_DESCENDANT)
-		{
-	    	lpq = TRUE;
-			g_string_append(xd->spath,"#/%");
-		}
-		else if(xd->path_q && lAxis==AXIS_ATTRIBUTE)
-		{
-			lpq = TRUE;
-			if(lprevCond->prevAxis==AXIS_DESCENDANT_OR_SELF)
-			{	
-				g_string_append(xd->spath,"#%/@%");
-			}
-			else
-				g_string_append(xd->spath,"#/@%");
-		}
-*/
 	}
 	else if(!strncmp(ntest,"processing-instruction(",23))
 	{
@@ -1326,14 +1490,26 @@ static int translateNodeTest(char *ntest, GString **query, XPathData *xd)
 		buf = g_strdup_printf("xml_path xp%d,xml_node xn%d",lcnt,lcnt);
 		stack_push(lqs->ST_From, buf, strlen(buf)+1);
 
-    	buf = g_strdup_printf("xp%d.pathexp %s '%s' AND "
-    						"xn%d.pathid = xp%d.pathid",
-    						lcnt,ope,spath,lcnt,lcnt
-    						);
+		if(hasDep)
+		{
+	    	buf = g_strdup_printf("xp%d.pathexp %s '%s' AND "
+	    						"xn%d.pathid = xp%d.pathid",
+	    						lcnt,ope,spath,lcnt,lcnt
+	    						);
+		}
+	    else
+	    {
+	    	buf = g_strdup_printf("xp%d.pathexp %s '%s' AND "
+	    						"xn%d.pathid = xp%d.pathid AND "
+	    						"xn%d.kind < 5",
+	    						lcnt,ope,spath,lcnt,lcnt,lcnt
+	    						);	    
+	    }
+	    	
     	stack_push(lqs->ST_Where, buf, strlen(buf)+1);
 
     	/* query setting */
-    	xd->distinct = FALSE;
+    	//xd->distinct = FALSE;
     	xd->fcount = 2;
 
     	/* cleen up */
@@ -1363,20 +1539,22 @@ static int translateNodeTest(char *ntest, GString **query, XPathData *xd)
 
 static int translatePredicate(char *argv, GString **query, XPathData *xd, int pcnt)
 {
-
 	static int s_cnt_pos;
 	char *predicate = strdup(argv);
 	char *p_ext;
-
 	char *ptr;
 	char *buf;
-	int i		= 0;
-	_Bool isDes = FALSE;
-
+	GString *clause;
+	
+	// for regex
 	regex_t preg;
 	char *pattern;
 	regmatch_t pmatch[8];
-
+	
+	// initializing
+	int i		= 0;
+	bool isDes = FALSE;
+	
     int lcnt	= xd->scount;
     QStack *lqs	= xd->qstack_val;
 
@@ -1467,7 +1645,7 @@ static int translatePredicate(char *argv, GString **query, XPathData *xd, int pc
 		{
 			// order by document order
     		g_string_sprintfa(subquery,
-    						"_xn1.dewey >= _xn2.dewey and "
+    						"_xn2.dewey >= _xn1.dewey and "
     						"_xn2.group_id = _xn1.group_id "
     						);
 		}
@@ -1475,7 +1653,7 @@ static int translatePredicate(char *argv, GString **query, XPathData *xd, int pc
 		{
 			// order by reverse document order
     		g_string_sprintfa(subquery,
-    						"_int_ge(_xn2.dewey,_xn1.dewey) "
+    						"_xn1.dewey >= _xn2.dewey "
     						);
 		}
 
@@ -1570,7 +1748,7 @@ static int translatePredicate(char *argv, GString **query, XPathData *xd, int pc
         	isDes = TRUE;
 	}
 
-	GString *clause = g_string_sized_new(300);
+	clause = g_string_sized_new(300);
 
 	p_ext = g_memdup(predicate,strlen(predicate)+50);
 	g_free(predicate);
@@ -1592,7 +1770,7 @@ static int translatePredicate(char *argv, GString **query, XPathData *xd, int pc
 	regfree(&preg);
 	if(i == 0)
 	{
-		_Bool like = FALSE;
+		bool like = FALSE;
 
 		if(strstr(p_ext,"%")!=NULL)
 			like = TRUE;
@@ -1635,11 +1813,11 @@ static int translatePredicate(char *argv, GString **query, XPathData *xd, int pc
 		}
 		else
 		{
-			_Bool negative = FALSE;
+			bool negative = FALSE;
 			char *left	= g_strdup_printf("%.*s",pmatch[1].rm_eo-pmatch[1].rm_so,&p_ext[pmatch[1].rm_so]);
 			char *ope	= g_strdup_printf("%.*s",pmatch[3].rm_eo-pmatch[3].rm_so,&p_ext[pmatch[3].rm_so]);
 			char *right = g_strdup_printf("%.*s",pmatch[5].rm_eo-pmatch[5].rm_so,&p_ext[pmatch[5].rm_so]);
-			_Bool no_att = strchr(left,'@')==NULL;
+			bool no_att = strchr(left,'@')==NULL;
 
 #ifdef __DEBUG__
 			fprintf(stderr,"predicate=%s\n",p_ext);
@@ -1801,8 +1979,8 @@ static int translatePredicate(char *argv, GString **query, XPathData *xd, int pc
 		char *left	= g_strdup_printf("%.*s",pmatch[1].rm_eo-pmatch[1].rm_so,&p_ext[pmatch[1].rm_so]);
 		char *ope	= g_strdup_printf("%.*s",pmatch[3].rm_eo-pmatch[3].rm_so,&p_ext[pmatch[3].rm_so]);
 		char *right	= g_strdup_printf("%.*s",pmatch[4].rm_eo-pmatch[4].rm_so,&p_ext[pmatch[4].rm_so]);
-		_Bool flg = (strchr(left,'@')==NULL) ? FALSE : TRUE;	// attribute,text(),comment(),pi().. 
-		//_Bool sv = (strstr(ope,"==") == NULL) ? FALSE : TRUE;	// string value
+		bool flg = (strchr(left,'@')==NULL) ? FALSE : TRUE;	// attribute,text(),comment(),pi().. 
+		//bool sv = (strstr(ope,"==") == NULL) ? FALSE : TRUE;	// string value
 
 		pattern = "^[^)(]+/(text|comment|processiong-instruction)+\\(\\)$";
 		/** 			   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~1
@@ -1906,7 +2084,7 @@ static int translateFuncCalls(char *, GString **, XPathData *){
 }//translateFuncCalls
 */
 
-static _Bool makeSQL(GString **query,XPathData *xd,int tmp_q,char mode)
+static bool makeSQL(GString **query,XPathData *xd,int tmp_q,char mode)
 {
 
 	char *buf=NULL;
@@ -1915,7 +2093,7 @@ static _Bool makeSQL(GString **query,XPathData *xd,int tmp_q,char mode)
 	int j=0;
 	int lcnt		= xd->scount;
 	int fcnt;
-	//_Bool rewrite;
+	//bool rewrite;
 	QStack *qs		= xd->qstack_val;
 	Stack *from		= qs->ST_From;
 	Stack *where	= qs->ST_Where;
@@ -1972,9 +2150,12 @@ static _Bool makeSQL(GString **query,XPathData *xd,int tmp_q,char mode)
 	{
 		g_string_sprintfa(*query,"SELECT%s"
 									"xn%d.docid,"
-			            			"xn%d.id",
+			            			"%sxn%d.id%s",
 			            		  xd->distinct ? " DISTINCT " : " ",
-		            			  lcnt-1,lcnt-1
+		            			  lcnt-1,
+		            			  (mode&RET_XML)==RET_XML ? "toxml(" : "",
+		            			  lcnt-1,
+		            			  (mode&RET_XML)==RET_XML ? ") as xml" : ""
 	            		);
 	}
 
@@ -2133,7 +2314,7 @@ static int StrReplace( char *target, const char *from, const char *to )
 	//int df, i, j, k;
 	char *limit;
 	char *tmp;
-	//_Bool flg = FALSE;
+	//bool flg = FALSE;
 
 	rcount = 0;
 	f_len = strlen(from);
@@ -2179,7 +2360,7 @@ static int StrReplace( char *target, const char *from, const char *to )
 /*
  *	StrisDigit -- Srting is Digit Number ?
  */
-static _Bool StrisDigit(char *s)
+static bool StrisDigit(char *s)
 {
 
 	while(*s){
