@@ -2,8 +2,6 @@
  - XPath.sql,v 1.1 2003/07/17 14:10:45 myui Exp
  */
 
--- BEGIN;
-
 --DROP FUNCTION xpath2sql(varchar);
 CREATE or REPLACE FUNCTION xpath2sql(varchar) RETURNS varchar
   AS '/usr/local/lib/xpsql/XPath.so'
@@ -35,8 +33,13 @@ CREATE or REPLACE FUNCTION vchar2real(VARCHAR) RETURNS REAL
   AS '/usr/local/lib/xpsql/XPath.so'
   LANGUAGE 'C' IMMUTABLE STRICT;
 
---DROP TYPE __retxpathexec;
---CREATE TYPE __retxpathexec AS (document integer, result varchar);
+DROP TYPE __retxpathexec;
+CREATE TYPE __retxpathexec AS (document integer, result varchar);
+
+CREATE or REPLACE FUNCTION xpath_evalx(varchar) RETURNS setof __retxpathexec
+  AS '/usr/local/lib/xpsql/XPath.so'
+  LANGUAGE 'C' STRICT;
+
 
 /*
 	-- Exsamples
@@ -45,7 +48,3 @@ select xpath2sql('/aaa/bbb[2][aaa="bbb"]');
 select * from xpath_eval('/aaa/bbb[2][aaa="bbb"]');
 
 */
-
--- END;
--- rollback;
--- commit;
\ No newline at end of file
