@@ -1,9 +1,23 @@
 /**
- * xpsql.c,v 1.1 2003/07/17 14:10:45 myui Exp
+ * xpsql.c,v 1.5 2003/07/17 14:10:45 myui Exp
  * 
  * ----------------------
  *	Copyright (c) 2003	Makoto Yui <yuin@bb.din.or.jp>
  *  All rights reserved.
+ *
+ *  This program is free software; you can redistribute it and/or
+ *  modify it under the terms of the GNU Lesser General Public License
+ *  as published by the Free Software Foundation; either version 2
+ *  of the License, or (at your option) any later version.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU Lesser General Public License for more details.
+ *
+ *  You should have received a copy of the GNU Lesser General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 /**
  * XMLPGSQL Copyright (C) 2001-2003 (Eurah) MediaFront Inc.
@@ -24,8 +38,9 @@
  * along with xpsql; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  */
- 
+
 //#define __DEBUG__
+//#define __INFO__				/* show infomations */
 
 #include <string.h>
 //#include <stdlib.h>				/* realloc */
@@ -35,14 +50,57 @@
 #include "lib/stringinfo.h"		/* for makeStringInfo() */
 #include "mb/pg_wchar.h"
 
+#define MAX_PLANS 10
+
+#ifndef INT4ARRAYOID
+#define INT4ARRAYOID 1007		/* see bootstrap.c, defined in there. */
+#endif
+
+/* Initial buffer size is setted to 8K(8092) byte */
+#define XBUFSIZE 8192
+
 #define _SPI_CONN() connected=(SPI_connect()==SPI_ERROR_CONNECT)
 #define _SPI_FIN() if(!connected)SPI_finish()
 
+#define R_NORMAL(r) \
+		switch(r) \
+		{ \
+			case 1: elog(NOTICE, "Status: SPI_OK_CONNECT");break; \
+			case 2: elog(NOTICE, "Status: SPI_OK_FINISH");break; \
+			case 3: elog(NOTICE, "Status: SPI_OK_FETCH");break; \
+			case 4: elog(NOTICE, "Status: SPI_OK_UTILITY");break; \
+			case 5: elog(NOTICE, "Status: SPI_OK_SELECT");break; \
+			case 6: elog(NOTICE, "Status: SPI_OK_SELINTO");break; \
+			case 7: elog(NOTICE, "Status: SPI_OK_INSERT");break; \
+			case 8: elog(NOTICE, "Status: SPI_OK_DELETE");break; \
+			case 9: elog(NOTICE, "Status: SPI_OK_UPDATE");break; \
+			case 10: elog(NOTICE, "Status: SPI_OK_CURSOR");break; \
+			default: elog(NOTICE, "Unknown status: %d", ret); \
+		}
+
+#define R_ERROR(r) \
+		switch (r) \
+		{ \
+			case -1: elog(ERROR, "[SPI_ERROR_CONNECT] %s", query);break; \
+			case -2: elog(ERROR, "[SPI_ERROR_COPY] %s", query);break; \
+			case -3: elog(ERROR, "[SPI_ERROR_OPUNKNOWN] %s", query);break; \
+			case -4: elog(ERROR, "[SPI_ERROR_UNCONNECTED] %s", query);break; \
+ 			case -5: elog(ERROR, "[SPI_ERROR_CURSOR] %s", query);break; \
+			case -6: elog(ERROR, "[SPI_ERROR_ARGUMENT] %s", query);break; \
+			case -7: elog(ERROR, "[SPI_ERROR_PARAM] %s", query);break; \
+			case -8: elog(ERROR, "[SPI_ERROR_TRANSACTION] %s", query);break; \
+			case -9: elog(ERROR, "[SPI_ERROR_NOATTRIBUTE] %s", query);break; \
+			case -10: elog(ERROR, "[SPI_ERROR_NOOUTFUNC] %s", query);break; \
+			case -11: elog(ERROR, "[SPI_ERROR_TYPUNKNOWN] %s", query);break; \
+			default: elog(ERROR, "[SPI_ERROR_OTHER] %s", query); \
+		}
+
 /**
  * exec_query
  *		-- internal function
  * 
- * Modified by Makoto Yui <yuin@bb.din.or.jp>
+ * Author: Makoto Yui <yuin@bb.din.or.jp>
+ * Copyright (c) 2003 Makoto Yui <yuin@bb.din.or.jp>
  */
 int
 exec_query(char *query)
@@ -52,41 +110,67 @@ exec_query(char *query)
 #ifdef __DEBUG__
 	elog(NOTICE, "Query:%s", query);
 #endif
-	if ((ret = SPI_exec(query, 0)) < 0) {
-		switch (ret) {
-		  case -1:
-			  elog(ERROR, "[SPI_ERROR_CONNECT] %s", query);break;
-		  case -2:
-			  elog(ERROR, "[SPI_ERROR_COPY] %s", query);break;
-		  case -3:
-			  elog(ERROR, "[SPI_ERROR_OPUNKNOWN] %s", query);break;
-		  case -4:
-			  elog(ERROR, "[SPI_ERROR_UNCONNECTED] %s", query);break;
-		  case -5:
-			  elog(ERROR, "[SPI_ERROR_CURSOR] %s", query);break;
-		  case -6:
-			  elog(ERROR, "[SPI_ERROR_ARGUMENT] %s", query);break;
-		  case -7:
-			  elog(ERROR, "[SPI_ERROR_PARAM] %s", query);break;
-		  case -8:
-			  elog(ERROR, "[SPI_ERROR_TRANSACTION] %s", query);break;
-		  case -9:
-			  elog(ERROR, "[SPI_ERROR_NOATTRIBUTE] %s", query);break;
-		  case -10:
-			  elog(ERROR, "[SPI_ERROR_NOOUTFUNC] %s", query);break;
-		  case -11:
-			  elog(ERROR, "[SPI_ERROR_TYPUNKNOWN] %s", query);break;
-		  default:
-			  elog(ERROR, "[SPI_ERROR_OTHER] %s", query);
-		}
+	if ((ret = SPI_exec(query,0)) < 0) {
+		R_ERROR(ret);
 	}
 #ifdef __DEBUG__
-	elog(NOTICE, "[Result:%d]", ret);
+	R_NORMAL(ret);
 #endif
 
 	return ret;
 }
 
+/**
+ * execp_query
+ * 
+ * @author Makoto Yui <yuin@bb.din.or.jp>
+ * @see spi.h
+ * @params qstr 
+ * 		query skeltion
+ * @params nargs
+ * 		number of input parameters ($1, $2, etc.)
+ * @params argtypes
+ * 		pointer to an array containing the OIDs of the data types of the parameters
+ * @params nulls
+ * 		 An array describing which parameters are null.
+ * @params ptype
+ * 		specify the type of plan.
+ */
+int
+execp_query(char *qstr, int nargs, Oid *argtypes, Datum *pvals, char *nulls, char ptype)
+{
+	EPlan *qplan;			/* prepared plan */
+	int stat;
+
+	if(ptype=='s')
+	{ /* for SELECT */
+		/* find query plan which stored in memory */
+		qplan = find_plan(qstr, &sPlans, &n_sPlans);
+	}
+	else if(ptype=='o' || ptype=='i' || ptype=='u' || ptype=='d')
+	{ /* for INSERT, UPDATE, DELETE */
+		qplan = find_plan(qstr, &oPlans, &n_oPlans);
+	}
+	//else /* take care */
+
+	/* if there is no plan, we have to prepare plan ... */
+	if (qplan->splan == NULL)
+	{
+#ifdef __INFO__
+		elog(NOTICE, "There is no allocated plan. create new plan.");
+#endif
+		create_plan(&qplan, qstr, nargs, argtypes);
+	}
+#ifdef __INFO__
+	else	elog(NOTICE, "Reuse already prepared plan.");
+#endif
+
+	/* Ok, execute prepared plan. */
+	stat = SPI_execp(qplan->splan, pvals, NULL, 0);
+	return stat;
+}
+
+
 /**
  * create_new_document:
  *		-- unspecified in DOM level 2, proprietary extention 
@@ -99,13 +183,12 @@ Datum
 create_new_document(PG_FUNCTION_ARGS)
 {
 	char *doc_file =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(0)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(0)));
 	int32 doc_id;
 	int32 node_id;
 	bool connected;
 	bool isnull;
 	Datum seqname;
-	//Datum tmp;
 	StringInfo str = makeStringInfo();
 
 #ifdef __DEBUG__
@@ -148,13 +231,8 @@ create_new_document(PG_FUNCTION_ARGS)
 						 node_id-1, KIND_ROOT, doc_id, doc_id,
 						 node_id, KIND_DUMMY, doc_id, doc_id);
 		exec_query(str->data);
-		initStringInfo(str);
-
-		//appendStringInfo(str, "INSERT INTO xml_path(pathid,pathexp) VALUES(0,'')");
-		//exec_query(str->data);
-		//initStringInfo(str);
-
-		pfree(str);
+		
+		pfree(str->data);
 		_SPI_FIN();
 		PG_RETURN_INT32(node_id-1);
 	}
@@ -171,7 +249,7 @@ Datum
 xml_tagid(PG_FUNCTION_ARGS)
 {
 	char *tag_name =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(0)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(0)));
 	int32 tag_id;
 	bool connected;
 	bool isnull;
@@ -204,9 +282,8 @@ xml_tagid(PG_FUNCTION_ARGS)
 		appendStringInfo(str, "INSERT INTO xml_tag VALUES(%d,'%s')",
 						 tag_id, tag_name);
 		exec_query(str->data);
-		//initStringInfo(str);
 	}
-	pfree(str);
+	pfree(str->data);
 	_SPI_FIN();
 	PG_RETURN_INT32(tag_id);
 }
@@ -222,7 +299,7 @@ Datum
 xml_attid(PG_FUNCTION_ARGS)
 {
 	char *att_name =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(0)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(0)));
 	int32 att_id;
 	bool connected;
 	bool isnull;
@@ -255,9 +332,8 @@ xml_attid(PG_FUNCTION_ARGS)
 		appendStringInfo(str, "INSERT INTO xml_attribute VALUES(%d,'%s')",
 						 att_id, att_name);
 		exec_query(str->data);
-		//initStringInfo(str);
 	}
-	pfree(str);
+	pfree(str->data);
 	_SPI_FIN();
 	PG_RETURN_INT32(att_id);
 }
@@ -274,7 +350,7 @@ Datum
 xml_nsid(PG_FUNCTION_ARGS)
 {
 	char *ns_name =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(0)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(0)));
 	int32 ns_id;
 	bool connected;
 	bool isnull;
@@ -308,9 +384,8 @@ xml_nsid(PG_FUNCTION_ARGS)
 		appendStringInfo(str, "INSERT INTO xml_namespace VALUES(%d,'%s')",
 						 ns_id, ns_name);
 		exec_query(str->data);
-		//initStringInfo(str);
 	}
-	pfree(str);
+	pfree(str->data);
 	_SPI_FIN();
 	PG_RETURN_INT32(ns_id);
 }
@@ -327,7 +402,7 @@ Datum
 xml_piid(PG_FUNCTION_ARGS)
 {
 	char *pi_name =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(0)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(0)));
 	int32 pi_id;
 	bool connected;
 	bool isnull;
@@ -359,10 +434,9 @@ xml_piid(PG_FUNCTION_ARGS)
 		appendStringInfo(str, "INSERT INTO xml_pi VALUES(%d,'%s')", 
 						 pi_id,pi_name);
 		exec_query(str->data);
-		//initStringInfo(str);
 	}
 
-	pfree(str);
+	pfree(str->data);
 	_SPI_FIN();
 	PG_RETURN_INT32(pi_id);
 }
@@ -378,7 +452,7 @@ Datum
 drop_document(PG_FUNCTION_ARGS)
 {
 	char *doc_file =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(0)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(0)));
 	int32 doc_id;
 	bool connected;
 	bool isnull;
@@ -402,13 +476,12 @@ drop_document(PG_FUNCTION_ARGS)
 						 "DELETE FROM xml_document WHERE docid=%d;DELETE FROM xml_node WHERE docid=%d",
 						 doc_id, doc_id);
 		exec_query(str->data);
-		//initStringInfo(str);
-		pfree(str);
+		pfree(str->data);
 		_SPI_FIN();
 		PG_RETURN_BOOL(TRUE);
 	}
 	else {
-		pfree(str);
+		pfree(str->data);
 		_SPI_FIN();
 		PG_RETURN_BOOL(FALSE);
 	}
@@ -425,7 +498,7 @@ Datum
 get_tagid(PG_FUNCTION_ARGS)
 {
 	char *tag_name =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(0)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(0)));
 	int32 tag_id;
 	bool connected;
 	bool isnull;
@@ -438,18 +511,17 @@ get_tagid(PG_FUNCTION_ARGS)
 	appendStringInfo(str, "SELECT tagid FROM xml_tag WHERE name='%s'",
 					 tag_name);
 	exec_query(str->data);
-	//initStringInfo(str);
+	pfree(str->data);
+
 	if (SPI_processed > 0) {
 		tag_id =
 			DatumGetInt32(SPI_getbinval
 						  (SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 1,
 						   &isnull));
-		pfree(str);
 		_SPI_FIN();
 		PG_RETURN_INT32(tag_id);
 	}
 	else {
-		pfree(str);
 		_SPI_FIN();
 		PG_RETURN_NULL();
 	}
@@ -466,7 +538,7 @@ Datum
 get_attid(PG_FUNCTION_ARGS)
 {
 	char *att_name =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(0)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(0)));
 	int32 att_id;
 	bool connected;
 	bool isnull;
@@ -481,18 +553,17 @@ get_attid(PG_FUNCTION_ARGS)
 					 "SELECT tagid FROM xml_attribute WHERE name='%s'",
 					 att_name);
 	exec_query(str->data);
-	//initStringInfo(str);
+	pfree(str->data);
+
 	if (SPI_processed > 0) {
 		att_id =
 			DatumGetInt32(SPI_getbinval
 						  (SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 1,
 						   &isnull));
-		pfree(str);
 		_SPI_FIN();
 		PG_RETURN_INT32(att_id);
 	}
 	else {
-		pfree(str);
 		_SPI_FIN();
 		PG_RETURN_NULL();
 	}
@@ -509,7 +580,7 @@ Datum
 get_nsid(PG_FUNCTION_ARGS)
 {
 	char *ns_name =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(0)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(0)));
 	int32 ns_id;
 	bool connected;
 	bool isnull;
@@ -524,18 +595,17 @@ get_nsid(PG_FUNCTION_ARGS)
 					 "SELECT tagid FROM xml_namespace WHERE name='%s'",
 					 ns_name);
 	exec_query(str->data);
-	//initStringInfo(str);
+	pfree(str->data);
+
 	if (SPI_processed > 0) {
 		ns_id =
 			DatumGetInt32(SPI_getbinval
 						  (SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 1,
 						   &isnull));
-		pfree(str);
 		_SPI_FIN();
 		PG_RETURN_INT32(ns_id);
 	}
 	else {
-		pfree(str);
 		_SPI_FIN();
 		PG_RETURN_NULL();
 	}
@@ -552,7 +622,7 @@ Datum
 get_piid(PG_FUNCTION_ARGS)
 {
 	char *pi_name =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(0)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(0)));
 	int32 pi_id;
 	bool connected;
 	bool isnull;
@@ -566,18 +636,17 @@ get_piid(PG_FUNCTION_ARGS)
 	appendStringInfo(str, "SELECT tagid FROM xml_pi WHERE name='%s'",
 					 pi_name);
 	exec_query(str->data);
-	//initStringInfo(str);
+	pfree(str->data);
+	
 	if (SPI_processed > 0) {
 		pi_id =
 			DatumGetInt32(SPI_getbinval
 						  (SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 1,
 						   &isnull));
-		pfree(str);
 		_SPI_FIN();
 		PG_RETURN_INT32(pi_id);
 	}
 	else {
-		pfree(str);
 		_SPI_FIN();
 		PG_RETURN_NULL();
 	}
@@ -612,8 +681,7 @@ Datum
 	createelement(PG_FUNCTION_ARGS)
 {
 	int32 doc_node=PG_GETARG_INT32(0);
-	char *tag_name=DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(1)));
-	//char *doc_id;
+	char *tag_name=DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(1)));
 	int doc_id;
 	int32 node_id;
 	int32 dummy_id;
@@ -661,7 +729,7 @@ Datum
 				);
 	exec_query(str->data);
 	
-	pfree(str);		
+	pfree(str->data);		
 	_SPI_FIN();
 	PG_RETURN_INT32(node_id);
 }
@@ -694,9 +762,8 @@ Datum
 	createprocessinginstruction(PG_FUNCTION_ARGS)
 {
 	int32 doc_node=PG_GETARG_INT32(0);
-	char *pi_name=DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(1)));
-	char *pi_data=DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(2)));
-	//char *doc_id;
+	char *pi_name=DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(1)));
+	char *pi_data=DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(2)));
 	int doc_id;
 	int32 node_id;
 	int32 dummy_id;
@@ -743,7 +810,7 @@ Datum
 				);
 	exec_query(str->data);
 
-	pfree(str);		
+	pfree(str->data);		
 	_SPI_FIN();
 	PG_RETURN_INT32(node_id);
 }
@@ -772,7 +839,7 @@ Datum
 	createtextnode(PG_FUNCTION_ARGS)
 {
 	int32 doc_node=PG_GETARG_INT32(0);
-	char *content=DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(1)));
+	char *content=DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(1)));
 	size_t len = strlen(content);
 	//char *doc_id;
 	int doc_id;
@@ -825,7 +892,7 @@ Datum
 					node_id);
 	exec_query(str->data);
 
-	pfree(str);		
+	pfree(str->data);		
 	_SPI_FIN();
 	PG_RETURN_INT32(node_id);
 }
@@ -854,7 +921,7 @@ Datum
 	createcomment(PG_FUNCTION_ARGS)
 {
 	int32 doc_node=PG_GETARG_INT32(0);
-	char *content=DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(1)));
+	char *content=DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(1)));
 	//char *doc_id;
 	int doc_id;
 	int32 node_id;
@@ -902,7 +969,7 @@ Datum
 				);
 	exec_query(str->data);
 	
-	pfree(str);		
+	pfree(str->data);		
 	_SPI_FIN();
 	PG_RETURN_INT32(node_id);
 }
@@ -920,7 +987,7 @@ new_childelement(PG_FUNCTION_ARGS)
 {
 	int32 parent_id = PG_GETARG_INT32(0);
 	char *tag_name =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(1)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(1)));
 	char *doc_id;
 	int32 node_id;
 	bool connected;
@@ -962,7 +1029,6 @@ new_childelement(PG_FUNCTION_ARGS)
 					 "VALUES(%d,'%d',%s,%d,xml_tagid('%s'))", 
 					 node_id, KIND_ELEMENT, doc_id, node_id, tag_name);
 	exec_query(str->data);
-	//initStringInfo(str);
 
 	/*
 	 * SET parent-child link
@@ -970,7 +1036,7 @@ new_childelement(PG_FUNCTION_ARGS)
 	DirectFunctionCall2(appendchild, Int32GetDatum(parent_id),
 						Int32GetDatum(node_id));
 
-	pfree(str);
+	pfree(str->data);
 	_SPI_FIN();
 	PG_RETURN_INT32(node_id);
 }
@@ -988,9 +1054,9 @@ new_childpi(PG_FUNCTION_ARGS)
 {
 	int32 parent_id = PG_GETARG_INT32(0);
 	char *pi_name =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(1)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(1)));
 	char *pi_data =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(2)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(2)));
 	char *doc_id;
 	int32 node_id;
 	bool connected;
@@ -1026,12 +1092,11 @@ new_childpi(PG_FUNCTION_ARGS)
 					 "VALUES(%d, '%d', %s, %d, xml_piid('%s'), '%s');",
 					 node_id, KIND_PI, doc_id, node_id, pi_name, pi_data);
 	exec_query(str->data);
-	//initStringInfo(str);
 
 	DirectFunctionCall2(appendchild, Int32GetDatum(parent_id),
 						Int32GetDatum(node_id));
 
-	pfree(str);
+	pfree(str->data);
 	_SPI_FIN();
 	PG_RETURN_INT32(node_id);
 }
@@ -1049,7 +1114,7 @@ new_childtext(PG_FUNCTION_ARGS)
 {
 	int32 parent_id = PG_GETARG_INT32(0);
 	char *content =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(1)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(1)));
 	size_t len = strlen(content);
 	char *doc_id;
 	int32 node_id;
@@ -1092,11 +1157,11 @@ new_childtext(PG_FUNCTION_ARGS)
 					 "VALUES(%d, '%d', %s, %d, '%s')", 
 					 node_id, KIND_TEXT, doc_id, node_id, content);
 	exec_query(str->data);
-	//initStringInfo(str);
+
 	DirectFunctionCall2(appendchild, Int32GetDatum(parent_id),
 						Int32GetDatum(node_id));
 
-	pfree(str);
+	pfree(str->data);
 	pfree(content);
 	_SPI_FIN();
 	PG_RETURN_INT32(node_id);
@@ -1115,7 +1180,7 @@ new_childcomment(PG_FUNCTION_ARGS)
 {
 	int32 parent_id = PG_GETARG_INT32(0);
 	char *content =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(1)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(1)));
 	char *doc_id;
 	int32 node_id;
 	bool connected;
@@ -1151,12 +1216,11 @@ new_childcomment(PG_FUNCTION_ARGS)
 					 "VALUES(%d, '%d', %s, %d, '%s')", 
 					 node_id, KIND_COMMENT, doc_id, node_id, content);
 	exec_query(str->data);
-	//initStringInfo(str);
 
 	DirectFunctionCall2(appendchild, Int32GetDatum(parent_id),
 						Int32GetDatum(node_id));
 
-	pfree(str);
+	pfree(str->data);
 	_SPI_FIN();
 	PG_RETURN_INT32(node_id);
 }
@@ -1182,9 +1246,9 @@ setattribute(PG_FUNCTION_ARGS)
 {
 	int32 parent_id = PG_GETARG_INT32(0);
 	char *att_name =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(1)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(1)));
 	char *att_value =
-		DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(2)));
+		DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(2)));
 	char *doc_id;
 	int32 node_id;
 	bool connected;
@@ -1240,10 +1304,9 @@ setattribute(PG_FUNCTION_ARGS)
 						 "UPDATE xml_node SET parent = %d WHERE id = %d",
 						 parent_id, node_id);
 		exec_query(str->data);
-		//initStringInfo(str);
 	}
 
-	pfree(str);
+	pfree(str->data);
 	_SPI_FIN();
 	PG_RETURN_INT32(node_id);
 }
@@ -1273,14 +1336,14 @@ setnamespace(PG_FUNCTION_ARGS)
 	if (fcinfo->nargs == 3)
 	{
 		default_ns=FALSE;
-		ns_name=DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(1)));
-		ns_value=DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(2)));	
+		ns_name=DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(1)));
+		ns_value=DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(2)));	
 	}
 	else if(fcinfo->nargs == 2)
 	{
 		default_ns=TRUE;
 		ns_name="xmlns";
-		ns_value=DatumGetCString(DirectFunctionCall1(textout, PG_GETARG_DATUM(1)));			
+		ns_value=DatumGetCString(DirectFunctionCall1(varcharout, PG_GETARG_DATUM(1)));			
 	}
 	else
 		PG_RETURN_NULL();
@@ -1336,10 +1399,9 @@ setnamespace(PG_FUNCTION_ARGS)
 						 "UPDATE xml_node SET parent = %d WHERE id = %d",
 						 parent_id, node_id);
 		exec_query(str->data);
-		//initStringInfo(str);
 	}
 
-	pfree(str);
+	pfree(str->data);
 	_SPI_FIN();
 	PG_RETURN_INT32(node_id);
 }
@@ -1429,7 +1491,6 @@ appendchild(PG_FUNCTION_ARGS)
 						 node_id, parent_node_id, node_id, KIND_ATTRIBUTE,
 						 KIND_NAMESPACE);
 		exec_query(str->data);
-		//initStringInfo(str);
 	}
 	else {
 		/*
@@ -1438,10 +1499,9 @@ appendchild(PG_FUNCTION_ARGS)
 		appendStringInfo(str, "UPDATE xml_node SET child = %d WHERE id = %d",
 						 node_id, parent_node_id);
 		exec_query(str->data);
-		//initStringInfo(str);
 	}
 
-	pfree(str);
+	pfree(str->data);
 	_SPI_FIN();
 	PG_RETURN_BOOL(TRUE);
 }
@@ -1483,27 +1543,15 @@ Datum
 	insertbefore(PG_FUNCTION_ARGS)
 {
 	int32 newChild;
-//	int32 thisNode;
 	int refChild;
 	int child_id;
 	char *parent_id,*ref_doc;
 	int kind_code;
 	bool connected;
-//	bool hierarchyCheck=FALSE;
 	StringInfo str = makeStringInfo();
 
-/*
-	if(fcinfo->nargs==3){
-		hierarchyCheck = TRUE;
-		thisNode = PG_GETARG_INT32(0);
-		newChild = PG_GETARG_INT32(1);
-		refChild = (int) PG_GETARG_INT32(2);
-	}
-	else{
-*/
-		newChild = PG_GETARG_INT32(0);
-		refChild = (int) PG_GETARG_INT32(1);	
-//	}
+	newChild = PG_GETARG_INT32(0);
+	refChild = (int) PG_GETARG_INT32(1);
 
 #ifdef __DEBUG__
 	elog(NOTICE,"[insertBefore] InsertNode:%d BeforeNode:%d",newChild,refChild);
@@ -1569,7 +1617,7 @@ Datum
 		exec_query(str->data);
 	}
 
-	pfree(str);
+	pfree(str->data);
 	_SPI_FIN();
 	PG_RETURN_BOOL(TRUE);
 }
@@ -1599,7 +1647,7 @@ stringvalue(PG_FUNCTION_ARGS)
 	char *d_order;
 	char *ret_text=NULL;
 	bool isnull;
-	struct varlena *p;
+	VarChar *result;
 	StringInfo str = makeStringInfo();
 
 #ifdef __DEBUG__
@@ -1636,10 +1684,10 @@ stringvalue(PG_FUNCTION_ARGS)
 									"xml_node "
 								"WHERE "
 									"kind='%d' AND "	//value IS NOT NULL AND
-									"_int_descendant(dewey,'%s') "
+									"dewey > '%s' AND dewey < nextSibling('%s') "
 								"ORDER BY dewey", 
-						   KIND_TEXT, d_order);
-		  elog(NOTICE, "%s", str->data);
+						   KIND_TEXT, d_order, d_order);
+		  //elog(NOTICE, "%s", str->data);
 		  exec_query(str->data);
 	      if (SPI_processed == 0) {
 			  _SPI_FIN();
@@ -1658,7 +1706,7 @@ stringvalue(PG_FUNCTION_ARGS)
 		  }
 		  ret_text = str->data;
 		  length = str->len;
-		  pfree(str);
+		  //pfree(str->data);
 
 		  break;
 	  case KIND_TEXT:
@@ -1669,7 +1717,7 @@ stringvalue(PG_FUNCTION_ARGS)
 		  appendStringInfo(str, "SELECT value FROM xml_node WHERE id = %d",
 						   node_id);
 		  exec_query(str->data);
-		  pfree(str);
+		  pfree(str->data);
 		  if (SPI_processed == 0) {
 			  _SPI_FIN();
 			  PG_RETURN_NULL();
@@ -1682,14 +1730,115 @@ stringvalue(PG_FUNCTION_ARGS)
 	// disconenct
 	_SPI_FIN();
 	
-	p = palloc(length + VARHDRSZ);
-	p->vl_len = length + VARHDRSZ;
-	memcpy(VARDATA(p), ret_text, length);
+	result = DatumGetVarCharP(DirectFunctionCall1(varcharin, CStringGetDatum(ret_text)));
 	pfree(ret_text);
 	
-	PG_RETURN_TEXT_P(p);
+	PG_RETURN_VARCHAR_P(result);
 }
 
+PG_FUNCTION_INFO_V1(toxml);
+Datum toxml(PG_FUNCTION_ARGS)
+{
+    Datum node_id = PG_GETARG_DATUM(0);
+	int i;
+    char *qstr;
+   	bool connected, isnull;
+    Datum d_start, d_nextsibling;
+	Datum pvals[10];
+	VarChar *result;
+	StringInfo str = makeStringInfo();
+	char *ret_text = NULL;
+	Oid argtypes1[1] = {INT4OID};
+	Oid argtypes[10] = {INT4ARRAYOID, INT4ARRAYOID,INT4ARRAYOID, INT4ARRAYOID, INT4ARRAYOID, 
+					   INT4ARRAYOID, INT4ARRAYOID,INT4ARRAYOID, INT4ARRAYOID, INT4ARRAYOID
+					};
+	/* have to allocate to accesable space */
+	Datum *pvals1 = (Datum *) palloc(sizeof(Datum));
+	pvals1[0] = node_id;
+
+#ifdef __DEBUG__
+    elog(NOTICE, "[toXML]");
+#endif
+
+    _SPI_CONN();
+    qstr = "SELECT dewey, nextSibling(dewey) FROM xml_node WHERE id = $1";
+    execp_query(qstr, 1, argtypes1, pvals1, NULL, 's');
+    pfree(pvals1);
+    if (SPI_processed == 0) {
+		_SPI_FIN();
+		PG_RETURN_NULL();
+    }
+
+    d_start = SPI_getbinval(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 1, &isnull);
+    d_nextsibling = SPI_getbinval(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 2, &isnull);
+	pvals[0] = pvals[2] = pvals[4] = pvals[6] = pvals[8] = d_start;
+	pvals[1] = pvals[3] = pvals[5] = pvals[7] = pvals[9] = d_nextsibling;
+	
+    qstr = "SELECT "
+				"xn.dewey as dewey, 1 as seq, '<' || xt.name as parsed_text "
+			"FROM xml_node xn, xml_tag xt "
+			"WHERE xn.tagid = xt.tagid AND xn.kind = '1' "
+				"AND (xn.dewey >= $1 AND xn.dewey < $2) "
+			"UNION ALL "
+			"SELECT " 
+				"xn.dewey, xn.kind - 3 as seq, CASE WHEN xn.kind = '5' THEN ' ' || xa.name || '=\"' || xn.value || '\"' ELSE ' ' || ns.name || '=\"' || xn.value || '\"' END "
+			"FROM xml_node xn LEFT JOIN xml_attribute xa USING(tagid) "
+				"LEFT JOIN xml_namespace ns USING(tagid) "
+			"WHERE (xn.kind = '5' OR xn.kind = '6') "
+				"AND (xn.dewey >= $3 AND xn.dewey < $4) "
+			"UNION ALL "
+			"SELECT "
+				"xn.dewey + 0, 4 as seq, '>' "
+			"FROM xml_node xn "
+			"WHERE xn.kind = '1' AND (xn.dewey >= $5 and xn.dewey < $6) "
+			"UNION ALL "
+			"SELECT "
+				"xn.dewey, 5 as seq, CASE WHEN xn.kind = '2' THEN xn.value WHEN xn.kind = '3' THEN '<!-- ' || xn.value || ' -->' ELSE '<?' || pi.name || xn.value || '?>' END "
+			"FROM xml_node xn LEFT JOIN xml_pi pi USING(tagid) "
+			"WHERE (xn.kind = '2' OR xn.kind = '3' OR xn.kind = '4') "
+				"AND (xn.dewey >= $7 AND xn.dewey < $8) "
+			"UNION ALL "
+			"SELECT "
+				"xn.dewey + 2147483647, 6 as seq, '</' || xt.name || '>' "
+			"FROM xml_node xn, xml_tag xt "
+			"WHERE xn.tagid = xt.tagid AND xn.kind = '1' "
+				"AND (xn.dewey >= $9 AND xn.dewey < $10) "
+			"ORDER BY dewey, seq";
+
+#ifdef __DEBUG__ 
+	elog(NOTICE, "%s", qstr);
+#endif
+    execp_query(qstr, 10, argtypes, pvals, NULL, 's');
+    if (SPI_processed == 0) {
+		_SPI_FIN();
+		PG_RETURN_NULL();
+    }
+    
+    /* get enough Buffer */
+	enlargeStringInfo(str, XBUFSIZE);
+
+    for (i = 0; i < SPI_processed; i++) {
+		ret_text = SPI_getvalue(SPI_tuptable->vals[i], SPI_tuptable->tupdesc, 3);
+		if (ret_text != NULL) {
+			/* if there is not enough buffer, the buffer space 
+			 * is doubled (repalloc) automatic */
+			appendStringInfo(str, ret_text);
+			pfree(ret_text);
+		}
+    }
+    
+	_SPI_FIN();	// disconenct
+
+	result = DatumGetVarCharP(DirectFunctionCall1
+							(varcharin, CStringGetDatum(str->data)));
+	pfree(str->data);
+	
+	if (result == NULL)
+		PG_RETURN_NULL();
+    PG_RETURN_VARCHAR_P(result);
+}
+
+
 /*
  *  StrReplace - replace string
  *  
@@ -1722,9 +1871,102 @@ StrReplace(char *ap, const char *bp, const char *cp)
 	return rc;
 }// streplace
 
+/* for caching of query plans, stolen from contrib/spi/\*.c */
+static EPlan *
+find_plan(char *ident, EPlan ** eplan, int *nplans)
+{
+	EPlan	   *newp;
+	int			i;
+
+#ifdef __INFO__
+		elog(NOTICE,"Show counters of prepared plans below.\n"
+					"Selecting..%d, Others..%d\n",
+			n_sPlans, n_oPlans);
+#endif
+
+	if (*nplans > MAX_PLANS)
+	{
+		/*
+		 * restrict the number of plans.
+		 * the oldest place is assigned newly.
+		 */
+		(*nplans)--;
+		pfree((*eplan)->ident);
+		newp = *eplan; //FIFO
+
+#ifdef __INFO__
+		elog(NOTICE,"There are too many plans. "
+					"The oldest plan is replaced to this new plan."); 
+#endif
+
+	}
+	else if (*nplans > 0)
+	{
+		// find a plan which matched to the specified label first.
+		for (i = 0; i < *nplans; i++)
+		{
+			if (strcmp((*eplan)[i].ident, ident) == 0)
+				break;
+		}
+		// found
+		if (i != *nplans)
+			return (*eplan + i);
+		// not found. expand the space and create new space for a plan.
+		*eplan = (EPlan *) realloc(*eplan, (i + 1) * sizeof(EPlan));
+		newp = *eplan + i;
+	}
+	else // there is no allocated space for plan.
+	{
+		newp = *eplan = (EPlan *) malloc(sizeof(EPlan));
+		(*nplans) = i = 0;
+	}
+
+	// set identifier.
+	newp->ident = (char *) malloc(strlen(ident) + 1);
+	strcpy(newp->ident, ident);
+	// initialize plan.
+	newp->splan = NULL;
+	
+	// increment plan counter
+	(*nplans)++;
+
+	return (newp); // return space for a plan which specified by the identifier
+}
+
+/*
+ * prepare plan and save it.
+ * 
+ * @return prepared plan
+ */
+static void 
+create_plan(EPlan **qplan, char *querystr, int nargs, Oid *argtypes)
+{
+	void	   *pplan;
+	
+	/* Prepare plan for query */
+	pplan = SPI_prepare(querystr, nargs, argtypes);
+	if (pplan == NULL)
+		elog(ERROR, "SPI_prepare returned %d", SPI_result);
+
+	/*
+	 * Remember that SPI_prepare places plan in current memory context
+	 * - so, we have to save plan in Top memory context for latter
+	 * use.
+	 */
+	pplan = SPI_saveplan(pplan);
+	if (pplan == NULL)
+		elog(ERROR, "SPI_saveplan returned %d", SPI_result);
+
+	(*qplan)->splan = (void *) malloc(sizeof(void *));
+	(*qplan)->splan = pplan;	// saving plan
+
+	return;
+}
+
+
 /**
  * indent option
- * -bad -bap -cdb -br -nce -npcs -nbc -lp -cpn -nfca -fc1 -i4 -ci4 -cli2 -d4 -ts4
+ * -bad -bap -cdb -br -nce -npcs -nbc -lp -nfca -fc1 -i4 -ci4 -cli2 -d4 -ts4
  * 
  * [c.h]
  * typedef signed int int32;               // Size 32bits 4byte (int4)
